(**************************************************************************

   	This file is part of the AtomicDensityMatrix package for Mathematica
    Copyright (C) 2004-2017 Simon M. Rochester <simon@rochesterscientific.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

**************************************************************************)


BeginPackage["AtomicDensityMatrix`RotatingWaveApproximation`", {"ProcessOptions`", "AtomicDensityMatrix`"}];

Begin["`Private`"];

(* ::Text::Bold:: *)
(*Should fix this so it doesn't require energy to be set!*)

(*Transformation of the Hamiltonian H under a passive time dependent unitary transformation U. Uses*)
(*H|\[Psi]\[RightAngleBracket]=I d/dt|\[Psi]\[RightAngleBracket] *)
(*and*)
(*|*)
(*Overscript[\[Psi], ~]\[RightAngleBracket]=U^\[Dagger]|\[Psi]\[RightAngleBracket]*)
(*for a passive transformation.*)

EffectiveHamiltonian[h_, u_, t_] := ECT[u].h.u - I ECT[u].D[u, t];

DropFastTerms[expr_, freq_, t_] :=
    Module[{freq1, expr1},
        {expr1, freq1} = Expand[
            {expr, EnsureList@freq} //.
                (f : Sin | Cos)[x_ /; ! FreeQ[x, t]] :> TrigToExp@f@ExpandAll@x //.
                    Exp[x_ /; ! FreeQ[x, t]] :> Exp@ExpandAll@x
        ];
        expr1 /. Flatten[{{Exp@ExpandAll[2 I # t + c_.] -> 0, Exp@ExpandAll[-2 I # t + c_.] -> 0} & /@ freq1}]
    ];

Options[RotatingWaveTransformMatrix] = {TimeVariable -> Automatic, ReturnFrequencyReplacements -> False, Method -> Automatic};

OptionProcessor[Method] = EnsureList@#&;

(RotatingWaveTransformMatrix[sys_?AtomicSystemQ,
    freqlist : {{_, _Rule | {__Rule}, ___} ..} | {Except[_Rule], _Rule | {__Rule}, ___}, OptionsPattern[]] :=
    Module[{timeVariable, method, returnFrequencyReplacements, shiftlist, freqreps, freqbackreps, freq, freqlist1,
        lowerpos, upperpos, lowerShifts, upperShifts, lowerUnfixed,
        upperUnfixed, Unfixed, transmat, coupling, diffFreq, i, j, k},
        (*initialize shiftlist and frequency replacement list*)
        shiftlist = ConstantArray[Unfixed, Length[sys]];
        freqreps = {}; freqbackreps = {};
        (*expand multiple couplings in one entry in freqlist to multiple entries, and delete third elements *)
        freqlist1 = If[MatchQ[freqlist, {_, _Rule | {__Rule}, ___}], {freqlist}, freqlist ];
        freqlist1 = freqlist1[[All, ;; 2]] /. {{f_, c : {__Rule}} :> Sequence @@ ({f, #} & /@ c)};
        (*loop over optical frequencies, defining transform-induced level shifts to eliminate frequencies when possible.
            If not possible, define replacements to write frequencies in terms of frequencies that have been eliminated.*)
        Do[
            freq = freqlist1[[i, 1]];
            coupling = freqlist1[[i, 2]];
            (* Find the positions of states specified to be coupled by this frequency *)
            {lowerpos, upperpos} = List @@ (StatePosition[sys, #] & /@ coupling);
            (* Find the previously specified level shifts for these states *)
            {lowerShifts, upperShifts} = Extract[shiftlist, #] & /@ {lowerpos, upperpos};
            (* Split out sublists of states that have different previously specified energy shifts*)
            lowerShifts = Transpose[{lowerShifts, lowerpos}];
            upperShifts = Transpose[{upperShifts, upperpos}];
            {lowerShifts, upperShifts} = Split[#, First[#1] === First[#2]&] & /@ {lowerShifts, upperShifts};
            {lowerShifts, lowerpos} = {lowerShifts[[All, All, 1]], lowerShifts[[All, All, 2]]};
            {upperShifts, upperpos} = {upperShifts[[All, All, 1]], upperShifts[[All, All, 2]]};
            (* Find groups of unfixed states *)
            {lowerUnfixed, upperUnfixed} = Map[AllSameQ[#, Unfixed] &, {lowerShifts, upperShifts}, {2}];
            (* Loop over the sublists of lower and upper states for this frequency, eliminating optical frequencies by shifting states where possible. *)
            Do[
                Which[
                    (* If neither upper or lower are fixed, and Method is not "ShiftLowerState", fix the lower at its current position and shift the upper down *)
                    lowerUnfixed[[j]] && upperUnfixed[[k]] && !MemberQ[method, "ShiftLowerState"],
                    shiftlist = ReplacePart[shiftlist, lowerpos[[j]] -> 0];
                    shiftlist = ReplacePart[shiftlist, upperpos[[k]] -> -freq],
                    (* If neither are fixed, and Method is "ShiftLowerState", fix the upper at its current position and shift the lower up *)
                    lowerUnfixed[[j]] && upperUnfixed[[k]] && MemberQ[method, "ShiftLowerState"],
                    shiftlist = ReplacePart[shiftlist, lowerpos[[j]] -> freq];
                    shiftlist = ReplacePart[shiftlist, upperpos[[k]] -> 0],
                    (* If the lower is fixed and the upper is unfixed, shift upper to compensate for lower shift, then shift down by optical frequency *)
                    ! lowerUnfixed[[j]] && upperUnfixed[[k]],
                    shiftlist = ReplacePart[shiftlist, upperpos[[k]] -> First[lowerShifts[[j]]] - freq],
                    (* If the lower is unfixed and the upper is fixed, shift lower to compensate for upper shift, then shift up by optical frequency *)
                    lowerUnfixed[[j]] && ! upperUnfixed[[k]],
                    shiftlist = ReplacePart[shiftlist, lowerpos[[j]] -> First[upperShifts[[k]]] + freq],
                    (* If both are fixed, find replacement rule and back substitution rule for writing the optical frequency in terms of the previously eliminated frequencies and difference frequency *)
                    ! lowerUnfixed[[j]] && ! upperUnfixed[[k]],
                    diffFreq = Unique[];
                    AppendTo[
                        freqreps,
                        {freq -> -(First[upperShifts[[k]]] - First[lowerShifts[[j]]]) + diffFreq, {Flatten@lowerpos[[j]], Flatten@upperpos[[k]]}}
                    ];
                    AppendTo[
                        freqbackreps,
                        {diffFreq -> freq + (First[upperShifts[[k]]] - First[lowerShifts[[j]]]), {Flatten@lowerpos[[j]], Flatten@upperpos[[k]]}}
                    ]
                ],
                {j, Length[lowerpos]}, {k, Length[upperpos]}
            ],
            {i, Length[freqlist1]}
        ];
        transmat = MatrixExp[I timeVariable DiagonalMatrix[shiftlist /. Unfixed -> 0]];
        If[returnFrequencyReplacements, {transmat, freqreps, freqbackreps}, transmat]
    ]) // ProcessOptions;

(RotatingWaveTransformMatrix[sys_?AtomicSystemQ, shiftlist : {__Rule}, OptionsPattern[]] :=
    Module[{reps, shifts, timeVariable},
        reps = (Alternatives @@ SelectStates[sys, #[[1]]] -> #[[2]] &) /@ shiftlist;
        shifts = sys /. reps /. AtomicState[__] -> 0;
        MatrixExp[I timeVariable DiagonalMatrix@shifts]
    ]) // ProcessOptions;

(RotatingWaveTransformMatrix[sys_?AtomicSystemQ, f : Except[_Rule], OptionsPattern[]] :=
    Module[{timeVariable, returnFrequencyReplacements, transMatrix},
        transMatrix = MatrixExp[-I timeVariable Chop@Replace[DiagonalMatrix[Energy /@ sys] /. 0. -> 0, x_ /; x =!= 0 -> f, {2}]];
        If[returnFrequencyReplacements,
            {transMatrix, {}, {}},
            transMatrix
        ]
    ]) // ProcessOptions;

Options[RotatingWaveApproximation] = {TransformMatrix -> Automatic, TimeVariable -> Automatic, Method -> Automatic};

(RotatingWaveApproximation[
    sys_?AtomicSystemQ, h_?OperatorQ,
    freqList : Except[_List | _Rule] | {Except[_Rule | _List] ..} | {Except[_Rule], _Rule | {__Rule}, ___} | {{Except[_Rule], _Rule | {__Rule}, ___} ..},
    opts : OptionsPattern[]
] := Module[
    {method, timeVariable, transformMatrix, rotatingHam, standardFreqList, freqReps, freqBackReps},
    (* Get transform matrix and replacement rules *)
    transformMatrix = transformMatrix /. Automatic :> RotatingWaveTransformMatrix[sys, freqList, ReturnFrequencyReplacements -> True, FilterRules@opts];
    {transformMatrix, freqReps, freqBackReps} = If[OperatorQ@transformMatrix, {transformMatrix, {}, {}}, transformMatrix];
    (* Apply transform matrix to Hamiltonian *)
    rotatingHam = EffectiveHamiltonian[h, transformMatrix, timeVariable];
    If[!MemberQ[method, "KeepFastTerms"],
        (* Write untransformed optical frequencies in terms of transformed optical frequencies and difference frequencies *)
        rotatingHam = Fold[ReplaceAllAt[#1, #2[[1]], Join[Tuples@#2[[2]], Tuples@Reverse@#2[[2]]]]&, rotatingHam, freqReps];
        (* Drop fast oscillating terms *)
        standardFreqList = Switch[freqList,
            Except[_List | _Rule],
            {{freqList}},
            {Except[_Rule | _List] ..},
            List /@ freqList,
            {Except[_Rule], _Rule | {__Rule}, ___},
            {freqList},
            {{Except[_Rule], _Rule | {__Rule}, ___} ..},
            freqList
        ] /. {x : Except[_List | _Rule], rules : {__Rule}, y___} :> {x, First@rules, y};
        rotatingHam = DropFastTerms[ExpandAll@rotatingHam, standardFreqList[[All, 1]], timeVariable];
        (* Substitute back for untransformed optical frequencies *)
        rotatingHam = Fold[ReplaceAllAt[#1, #2[[1]], Join[Tuples@#2[[2]], Tuples@Reverse@#2[[2]]]]&, rotatingHam, freqBackReps];
        (* Write all frequencies in terms of detunings, if specified *)
        rotatingHam = rotatingHam /. Cases[
            standardFreqList,
            {freq : Except[_List | _Rule], crit1_ -> crit2_, delta : Except[_List | _Rule]} :> (freq -> Energy[SelectState[sys, crit2]] - Energy[SelectState[sys, crit1]] + delta)
        ]
    ];
    rotatingHam
]) // ProcessOptions;

RotatingWaveApproximation[as_?AtomicSystemQ, reg_ExperimentalRegion, h_?OperatorQ, opts : OptionsPattern[]] /;
    DefinedQ[reg, OpticalParameters] :=
        Module[{ops = OpticalParameters@reg},
            RotatingWaveApproximation[as, h, standardizeField[ops, {}, {}][[1, 1]], opts]
        ];

RotatingWaveApproximation[as_?AtomicSystemQ, reg_ExperimentalRegion, h_?OperatorQ, OptionsPattern[]] /;
    !DefinedQ[reg, OpticalParameters] := h;

RotatingWaveApproximation[as_?AtomicSystemQ, es : {_ExperimentalRegion..}, h : {_?OperatorQ..}, opts : OptionsPattern[]] :=
    MapThread[RotatingWaveApproximation[as, #1, #2, opts]&, {es, h}];

End[];

EndPackage[];

